"""
This module defines the User and IPAUser classes which can be used
to represent system and IPA users.

The classes contain the usual properties that defines a user, like username,
password, smart card pin, etc.

The classes implement add_user and delete_user methods which can be used to
create or remove a specified user in the system or in the specified IPA server.
"""
from shutil import rmtree

import json
import pwd
import python_freeipa
from pathlib import Path, PosixPath

from SCAutolib import run, logger, LIB_DUMP_USERS
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.models import card as card_model
from SCAutolib.models.CA import IPAServerCA
from SCAutolib.models.file import OpensslCnf


class BaseUser:
    username: str = None
    password: str = None
    pin: str = None
    dump_file: Path = None
    _cnf: OpensslCnf = None
    _key: Path = None
    _cert: Path = None
    card_dir: Path = None
    _card: card_model.Card = None
    local: bool = None

    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.dump_file = LIB_DUMP_USERS.joinpath(f"{self.username}.json")

    def to_dict(self):
        # Retype patlib.Path object to str
        d = {k: str(v) if type(v) in (PosixPath, Path) else v
             for k, v in self.__dict__.items()}

        if self._card and isinstance(self._card, card_model.VirtualCard):
            d.pop("_card")
            d["card"] = str(self._card.dump_file)

        return d

    @staticmethod
    def load(json_file, **kwargs):
        """
        Load values from user's JSON file to corresponding user object.

        :param json_file: path to JSON file to read from
        :type json_file: pathlib.Path
        :param kwargs: dictionary of additional values needed to initialise the
            object
        :return: object of local or IPA user
        :rtype: SCAutolib.models.user.User or SCAutolib.models.user.IPAUser
        """
        with json_file.open("r") as f:
            cnt = json.load(f)

        if "card_dir" not in cnt:
            user = BaseUser(username=cnt["username"], password=cnt["password"])
        elif cnt["local"]:
            user = User(local=cnt["local"],
                        username=cnt["username"],
                        card_dir=Path(cnt["card_dir"]),
                        password=cnt["password"],
                        pin=cnt["pin"],
                        key=cnt["_key"],
                        cert=cnt["_cert"])
        else:
            if "ipa_server" not in kwargs:
                raise SCAutolibException("IPA Server object does not provided. "
                                         "Can't load IPA user.")

            user = IPAUser(ipa_server=kwargs["ipa_server"],
                           local=cnt["local"],
                           username=cnt["username"],
                           card_dir=Path(cnt["card_dir"]),
                           password=cnt["password"],
                           pin=cnt["pin"],
                           key=cnt["_key"],
                           cert=cnt["_cert"])
        logger.debug(f"User {user.__class__} is loaded: {user.__dict__}")
        if "card" in cnt:
            return user, Path(cnt["card"])
        return user

    def add_user(self):
        """
        Add user to the local system with `useradd` bash command and set
        password for created user.
        :return:
        """
        try:
            pwd.getpwnam(self.username)
            msg = f"User {self.username} already exists on this " \
                  f"machine. Username should be unique to avoid " \
                  f"future problems with collisions"
            logger.critical(msg)
            raise SCAutolibException(msg)
        except KeyError:
            logger.debug(f"Creating new user {self.username}")
            cmd = ['useradd', '-m', self.username]
            run(cmd, check=True)
            cmd = ["passwd", self.username, "--stdin"]
            run(cmd, input=self.password)
            logger.info(f"User {self.username} is present on the system")


class User(BaseUser):
    """
    Generic class to represent system users.
    """

    def __init__(self, username: str, password: str, pin: str,
                 cnf: Path = None, key: Path = None, cert: Path = None,
                 card_dir: Path = None, local: bool = True):

        """
        :param username: Username for the system user
        :type username: str
        :param password: Password for the system user
        :type password: str
        :param pin: Smart card pin for the system user
        :type pin: str
        :param cnf: CNF file to be associated with the user
        :type cnf: Path
        :param key: Key to be associated with the user
        :type key: Path
        :param cert: Certificate to be associated with the user.
        :type cert: Path
        :param card_dir: Directory for the card. If None, standard
            home directory would be used (/home/<username>)
        :type card_dir: Path
        """

        self.username = username
        self.password = password
        self.pin = pin
        self.dump_file = LIB_DUMP_USERS.joinpath(f"{self.username}.json")
        self._cnf = cnf
        self.card_dir = card_dir if card_dir is not None \
            else Path("/home", self.username)
        self._key = key if key else self.card_dir.joinpath(f"key-{username}.pem")
        self._cert = cert \
            if cert else self.card_dir.joinpath(f"cert-{username}.pem")
        self.local = local

    @property
    def card(self):
        return self._card

    @card.setter
    def card(self, card: card_model.Card):
        if self._card:
            logger.error("Delete the existing card before adding a new one.")
            raise ValueError("A card is already assigned to this user")
        self._card = card

    @card.deleter
    def card(self):
        logger.info(f"Deleting the existing card from {self.username}")
        self._card = None

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, key: Path):
        logger.warning("Make sure to remove the existing key/cert "
                       "pairs before adding a new one.")
        self._key = key
        if self.card:
            self.card._private_key = key

    @key.deleter
    def key(self):
        logger.info("Deleting the current user key.")
        self._key = None

    @property
    def cert(self):
        return self._cert

    @cert.setter
    def cert(self, cert: Path):
        logger.warning("Make sure to remove the existing key/cert "
                       "pairs before adding a new one.")
        self._cert = cert
        if self.card:
            self.card._cert = cert

    @cert.deleter
    def cert(self):
        logger.info("Deleting the current user cert.")
        self._cert = None

    @property
    def cnf(self):
        return self._cnf

    @cnf.setter
    def cnf(self, cnf: Path):
        if self._cnf:
            logger.warning("Overriding current CNF file.")
        self._cnf = cnf

    @cnf.deleter
    def cnf(self):
        logger.info("Removing current CNF file.")
        self._cnf = None

    def delete_user(self):
        """
        Deletes the user and the content of user's card directory

        Note: card directory would be recursively deleted with a directory
        by itself.
        """
        try:
            pwd.getpwnam(self.username)
            logger.info(f"Deleting the user {self.username}")
            run(['userdel', '-f', self.username], check=True)
            if self.card_dir.exists():
                rmtree(self.card_dir)
                logger.debug("User's card directory "
                             f"{str(self.card_dir)} is removed")
        except KeyError:
            pass
        logger.info(f"User {self.username} is not present on the system")

    def gen_csr(self):
        """
        Method for generating local user specific CSR file that would be sent to
        the local CA for generating the certificate. CSR is generated using
        `openssl` command based on template CNF file.
        """
        csr_path = self.card_dir.joinpath(f"csr-{self.username}.csr")
        cmd = ["openssl", "req", "-new", "-nodes", "-key", self._key,
               "-reqexts", "req_exts", "-config", self._cnf, "-out", csr_path]
        run(cmd)
        return csr_path


class IPAUser(User):
    """
    This class is used to represent an IPA user.
    """
    default_password = "redhat"

    def __init__(self, ipa_server: IPAServerCA, *args, **kwargs):
        """
        Class for IPA user. IPA client should be configured first before
        creating an IPA user through this class.

        :param ipa_server: IPAServerCA object which provides the ipa hostname
        :type ipa_server: IPAServerCA
        :param username: Username for the system user
        :type username: str
        :param password: Password for the system user
        :type password: str
        :param pin: Smart card pin for the system user
        :type pin: str
        :param cnf: CNF file to be associated with the user
        :type cnf: Path
        :param key: Key to be associated with the user
        :type key: Path
        :param cert: Certificate to be associated with the user.
        :type cert: Path
        """

        super().__init__(*args, **kwargs)
        self._meta_client = ipa_server.meta_client
        self._ipa_hostname = ipa_server.ipa_server_hostname

    def to_dict(self):
        d = super().to_dict()
        d.pop("_meta_client")
        d.pop("_ipa_hostname")
        return d

    def add_user(self):
        """
        Adds IPA user to IPA server.
        """
        try:
            r = self._meta_client.user_add(self.username, self.username,
                                           self.username, self.username,
                                           o_userpassword=self.default_password)
            logger.debug(r)

            # To avoid forcing IPA server to change the password on first login
            # we changing it through the client
            client = python_freeipa.client.Client(self._ipa_hostname,
                                                  verify_ssl=False)
            client.change_password(self.username, self.password,
                                   self.default_password)
            logger.info(f"User {self.username} is added to the IPA server")
        except python_freeipa.exceptions.DuplicateEntry:
            msg = f"User {self.username} already exists on the " \
                  f"IPA server. Username should be unique to avoid " \
                  f"future problems with collisions"
            logger.critical(msg)
            raise SCAutolibException(msg)

    def delete_user(self):
        """
        Deletes the user and user's card directory.

        Note: card directory would be recursively deleted with a directory
        by itself.
        """
        try:
            r = self._meta_client.user_del(self.username)["result"]
            logger.info(f"User {self.username} is removed from the IPA server")
            logger.debug(r)
        except python_freeipa.exceptions.NotFound:
            pass
        if self.card_dir.exists():
            rmtree(self.card_dir)
            logger.info(f"User  {self.username} directory is removed.")

    def gen_csr(self):
        """
        Method for generating IPA user specific CSR file that would be sent to
        the IPA server for generating the certificate. CSR is generated using
        `openssl` command.
        """
        if not self._key:
            raise SCAutolibException("Can't generate CSR because private key "
                                     "is not set")
        csr_path = self.card_dir.joinpath(f"csr-{self.username}.csr")
        cmd = ["openssl", "req", "-new", "-days", "365",
               "-nodes", "-key", self._key, "-out",
               str(csr_path), "-subj", f"/CN={self.username}"]
        run(cmd)
        return csr_path



import json
from pathlib import Path
from schema import Schema, Use
from shutil import rmtree
from typing import Union

from SCAutolib import exceptions, schema_cas, schema_user
from SCAutolib import (logger, run, LIB_DIR, LIB_BACKUP, LIB_DUMP,
                       LIB_DUMP_USERS, LIB_DUMP_CAS, LIB_DUMP_CARDS)
from SCAutolib.models import CA, file, user, card, authselect as auth
from SCAutolib.models.file import File
from SCAutolib.utils import (OSVersion, _check_selinux, _gen_private_key,
                             _get_os_version, _install_packages,
                             _check_packages, dump_to_json, local_ca_factory)


class Controller:
    authselect: auth.Authselect = auth.Authselect()
    sssd_conf: file.SSSDConf = file.SSSDConf()
    lib_conf: dict = None
    _lib_conf_path: Path = None
    local_ca: CA.LocalCA = None
    ipa_ca: CA.IPAServerCA = None
    users: [user.User] = None

    @property
    def conf_path(self):
        return self._lib_conf_path

    def __init__(self, config: Union[Path, str], params: {} = None):
        """
        Constructor will parse and check input configuration file. If some
        required fields in the configuration are missing, CLI parameters
        would be checked if missing values are there. If not, an exception
        would be raised. After parsing the configuration file and filling
        internal values of the Controller object, other related objects (users,
        cards, CAs, Authselect, etc.) would be initialised, but any real action
        that would affect the system
        wouldn't be made.

        :param config: Path to configuration file with metadata for testing.
        :type config: pathlib.Path or str
        :param params: Parameters from CLI
        :type params: dict
        :return:
        """

        # Check params

        # Parse config file
        self._lib_conf_path = config.absolute() if isinstance(config, Path) \
            else Path(config).absolute()

        with self._lib_conf_path.open("r") as f:
            tmp_conf = json.load(f)
            if tmp_conf is None:
                raise exceptions.SCAutolibException(
                    "Data are not loaded correctly.")
        self.lib_conf = self._validate_configuration(tmp_conf, params)
        self.users = []
        for d in (LIB_DIR, LIB_BACKUP, LIB_DUMP, LIB_DUMP_USERS, LIB_DUMP_CAS,
                  LIB_DUMP_CARDS):
            d.mkdir(exist_ok=True)

    def prepare(self, force: bool, gdm: bool, install_missing: bool):
        """
        Method for setting up whole system based on configuration file and
        CLI commands
        """
        self.setup_system(install_missing, gdm)

        # In this method not having section for local CA and/or for IPA CA is OK
        try:
            self.setup_local_ca(force=force)
        except exceptions.SCAutolibWrongConfig as e:
            logger.info(e)
        try:
            self.setup_ipa_client(force=force)
        except exceptions.SCAutolibWrongConfig as e:
            logger.info(e)
        for usr in self.lib_conf["users"]:
            u = self.setup_user(usr, force=force)
            self.enroll_card(u)

    def setup_system(self, install_missing: bool, gdm: bool):
        """
        Do general system setup meaning package installation based on
        specifications in the configuration file, SSSD configuration,
        configurations for virtual smart cards, etc.

        :param install_missing: If True, all missing packages would be
            installed
        :type install_missing: bool
        :param gdm: If True, GDM package would be installed
        :type gdm: bool
        :return:
        """
        for d in (LIB_DIR, LIB_BACKUP, LIB_DUMP, LIB_DUMP_USERS, LIB_DUMP_CAS,
                  LIB_DUMP_CARDS):
            d.mkdir(exist_ok=True)

        packages = ["opensc", "httpd", "sssd", "sssd-tools", "gnutls-utils"]
        if gdm:
            packages.append("gdm")

        # Prepare for virtual cards
        if "virtual" in [u["card_type"] for u in self.lib_conf["users"]]:
            packages += ["pcsc-lite-ccid", "pcsc-lite", "virt_cacard",
                         "vpcd", "softhsm"]
            run("dnf -y copr enable jjelen/vsmartcard")

        # Add IPA packages if needed
        if not all([u["local"] for u in self.lib_conf["users"]]):
            packages += self._general_steps_for_ipa()

        # Check for installed packages
        missing = _check_packages(packages)
        if install_missing and missing:
            _install_packages(missing)
        elif missing:
            msg = "Can't continue. Some packages are missing: " \
                  f"{', '.join(missing)}"
            logger.critical(msg)
            raise exceptions.SCAutolibException(msg)

        run(['dnf', 'groupinstall', "Smart Card Support", '-y'])
        logger.debug("Smart Card Support group in installed.")

        self.sssd_conf.create()
        self.sssd_conf.save()
        self._general_steps_for_virtual_sc()

        base_user = user.BaseUser("base-user", "redhat")
        base_user.add_user()
        dump_to_json(base_user)
        dump_to_json(user.BaseUser(username="root",
                                   password=self.lib_conf["root_passwd"]))

    def setup_local_ca(self, force: bool = False):
        """
        Setup local CA based on configuration from the configuration file. All
        necessary file for this operation (e.g. CNF file for self-signed root
        certificate) would be created along the way.

        :param force: If local CA already exists in given directory, specifies
            if it should be overwritten
        :type force: bool
        :raises: SCAutolib.exceptions.SCAutolibWrongConfig
        """

        if "local_ca" not in self.lib_conf["ca"]:
            msg = "Section for local CA is not found in the configuration file"
            raise exceptions.SCAutolibWrongConfig(msg)

        ca_dir: Path = self.lib_conf["ca"]["local_ca"]["dir"]
        self.local_ca = local_ca_factory(ca_dir, force)
        logger.info(f"Local CA is configured in {ca_dir}")

        dump_to_json(self.local_ca)

    def setup_ipa_client(self, force: bool = False):
        """
        Configure IPA client for given IPA server on current host. IPA server
        should be already up and running for correct configuration of the IPA
        client

        :param force: If IPA Client is already configured on the system,
            specifies if it should be removed before configuring a new client.
        :type force: bool
        :raises: SCAutolib.exceptions.SCAutolibWrongConfig
        """
        if "ipa" not in self.lib_conf["ca"]:
            msg = "Section for IPA is not found in the configuration file"
            raise exceptions.SCAutolibWrongConfig(msg)
        self.ipa_ca = CA.IPAServerCA(**self.lib_conf["ca"]["ipa"])

        if self.ipa_ca.is_installed:
            logger.warning("IPA client is already configured on this system.")
            if not force:
                logger.info("Set force argument to True if you want to remove "
                            "previous installation.")
                return
            self.ipa_ca.cleanup()
        else:
            logger.info("IPA client is not configured on the system")
        self.ipa_ca.setup()
        self.sssd_conf.update_default_content()
        self.sssd_conf.set(key="domains",
                           value=f"shadowutils, {self.ipa_ca.domain}",
                           section="sssd")
        dump_to_json(self.ipa_ca)

    def setup_user(self, user_dict: dict, force: bool = False):
        """
        Configure the user on the specified system (local machine/CA). The user
        would be configured along with the card based on configurations.

        :param force: specify if the user should be re-created with its
            card directory
        :type force: bool
        :param user_dict: set of values to initialise the user
        :type user_dict: dict
        :return: the user object
        """
        new_user = None
        card_dir: Path = user_dict["card_dir"]
        # Card dir is home dir of the user home directory
        if force and card_dir.exists():
            rmtree(card_dir)

        if user_dict["local"]:
            new_user = user.User(username=user_dict["name"],
                                 pin=user_dict["pin"],
                                 password=user_dict["passwd"],
                                 card_dir=card_dir,
                                 cert=user_dict["cert"], key=user_dict["key"],
                                 local=True)
            if force:
                if new_user.cert and new_user.cert.exists():
                    self.local_ca.revoke_cert(new_user.cert)
                new_user.delete_user()
            user_dict["card_dir"].mkdir(exist_ok=True)
            new_user.add_user()

            csr_path = new_user.card_dir.joinpath(f"csr-{new_user.username}.csr")
            cnf = file.OpensslCnf(filepath=csr_path, conf_type="user",
                                  replace=new_user.username)
            cnf.create()
            cnf.save()

            new_user.cnf = cnf.path
            self.sssd_conf.set(
                section=f"certmap/shadowutils/{new_user.username}",
                key="matchrule",
                value=f"<SUBJECT>.*CN={new_user.username}.*")
            self.sssd_conf.save()
            self.sssd_conf.update_default_content()
            run(["systemctl", "restart", "sssd"])
            logger.debug(f"Match rule for user {new_user.username} is added "
                         f"to /etc/sssd/sssd.conf")
        else:
            if self.ipa_ca is None:
                msg = "Can't proceed in configuration of IPA user because no " \
                      "IPA Client is configured"
                raise exceptions.SCAutolibException(msg)
            new_user = user.IPAUser(ipa_server=self.ipa_ca,
                                    username=user_dict["name"],
                                    pin=user_dict["pin"],
                                    password=user_dict["passwd"],
                                    card_dir=card_dir,
                                    cert=user_dict["cert"],
                                    key=user_dict["key"],
                                    local=False)
            if force:
                if new_user.cert and new_user.cert.exists():
                    self.ipa_ca.revoke_cert(new_user.cert)
                new_user.delete_user()
            user_dict["card_dir"].mkdir(exist_ok=True)

            new_user.add_user()

        if user_dict["card_type"] == "virtual":
            hsm_conf = file.SoftHSM2Conf(
                new_user.card_dir.joinpath("sofhtsm2.conf"),
                new_user.card_dir)
            hsm_conf.create()
            hsm_conf.save()

            new_user.card = card.VirtualCard(new_user,
                                             softhsm2_conf=hsm_conf.path)
        else:
            raise NotImplementedError("Other card type than 'virtual' does"
                                      "not supported yet")

        new_user.card.create()
        self.users.append(new_user)

        dump_to_json(new_user)

        return new_user

    def enroll_card(self, user_: user.User, force: bool = False):
        """
        Enroll the card of a given user with configured CA. If private key
        and/or the certificate do not exists, new one's would be requested
        from corresponding CA.

        :param force:
        :param user_: User with a card to be enrolled.
        """
        logger.debug(f"Starting enrollment of the card for user "
                     f"{user_.username}")
        if not user_.card:
            raise exceptions.SCAutolibException(
                f"Card for the user {user_.username} is not initialized")

        if not user_.key.exists():
            _gen_private_key(user_.key)

        if not user_.cert.exists():
            # Creating a new private key makes sense only if the certificate
            # doesn't exist yet

            csr = user_.gen_csr()

            ca = self.ipa_ca \
                if isinstance(user_, user.IPAUser) else self.local_ca
            user_.cert = ca.request_cert(csr, user_.username, user_.cert)

        user_.card.enroll()

        dump_to_json(user_.card)
        # Update the dump for current user to link it with the card
        dump_to_json(user_)

    def cleanup(self):
        """
        Clean the system after setup. This method restores the SSSD config file,
        deletes created users with cards, remove CA's (local and/or IPA Client)
        """
        self.sssd_conf.clean()

        for user_file in LIB_DUMP_USERS.iterdir():
            usr = user.BaseUser.load(user_file)
            usr.delete()
            card_file = LIB_DUMP_CARDS.joinpath(f"card-{usr.username}.json")
            if card_file.exists():
                card.Card.load(card_file, user=user).delete()

        if self.local_ca:
            self.local_ca.cleanup()
        if self.ipa_ca:
            self.ipa_ca.cleanup()

        self.authselect.cleanup()

    @staticmethod
    def _validate_configuration(conf: dict, params: {} = None) -> dict:
        """
        Validate schema of the configuration file. If some value is not present
        in the config file, this value would be looked in the CLI parameters

        :param conf: Configuration to be parsed (e.g. data loaded from
            JSON file)
        :type conf: dict
        :param params: CLI arguments
        :type params: dict
        :return: dictionary with parsed values from conf and params attributes.
            All values are retyped to specified type.
        :rtype: dict
        """
        # FIXME: any schema requires all values to be in the config file, and
        #  only IP address of IPA server is accepted from CLI arguments.
        #  Add loading of the values from params dict
        # IP regex
        # Specify validation schema for CAs

        # Specify general schema for whole config file
        schema = Schema({"root_passwd": Use(str),
                         "ca": schema_cas,
                         "users": [schema_user]})

        return schema.validate(conf)

    @staticmethod
    def _general_steps_for_virtual_sc():
        """
        Prepare the system for virtual smart card. Preparation means to
        configure pcscd service and opensc module to work correctly
        with virtual smart card. Also, repository for installing virt_cacard
        package is added in this method.
        """

        _check_selinux()

        pcscd_service = File("/usr/lib/systemd/system/pcscd.service")
        pcscd_service.backup()
        exec_start = pcscd_service.get(section="Service", key="ExecStart")
        if "--auto-exit" in exec_start:
            exec_start = exec_start.replace("--auto-exit", "")
            pcscd_service.set(section="Service", key="ExecStart",
                              value=exec_start)
            pcscd_service.save()

        opensc_module = File("/usr/share/p11-kit/modules/opensc.module")
        opensc_module.backup()
        try:
            opensc_module.get("disable-in", separator=":")
        except exceptions.SCAutolibException:
            logger.warning("OpenSC module does not have option 'disable-in: "
                           "virt_cacard' set")
            opensc_module.set(key="disable-in", value="virt_cacard",
                              separator=": ")
            opensc_module.save()

        run(['systemctl', 'stop', 'pcscd.service', 'pcscd.socket', 'sssd'])
        rmtree("/var/lib/sss/mc/*", ignore_errors=True)
        rmtree("/var/lib/sss/db/*", ignore_errors=True)
        logger.debug(
            "Directories /var/lib/sss/mc/ and /var/lib/sss/db/ removed")

        run("systemctl daemon-reload")
        run("systemctl restart pcscd sssd")

        logger.debug("Copr repo for virt_cacard is enabled")

    @staticmethod
    def _general_steps_for_ipa():
        """
        General system preparation for installing IPA client on RHEL/Fedora

        :return: name of the IPA client package for current Linux
        """
        os_version = _get_os_version()
        if os_version not in (OSVersion.RHEL_9, OSVersion.CentOS_9):
            run("dnf module enable -y idm:DL1")
            run("dnf install @idm:DL1 -y")
            logger.debug("idm:DL1 module is installed")

        if os_version == OSVersion.Fedora:
            return ["freeipa-client"]
        else:
            return ["ipa-client"]

    def get_user_dict(self, name):
        """
        Get user dictionary from the config file.

        :param name: name of the user
        :type name: str
        :return: user dictionary
        :rtype: dict
        """
        for user_dict in self.lib_conf["users"]:
            if user_dict["name"] == name:
                return user_dict
        raise exceptions.SCAutolibMissingUserConfig(name)

    def init_ca(self, local: bool = False):
        """
        Initialize CA.

        :param local: if True, local CA is initialized, otherwise IPA
        :type local: bool
        """
        if local:
            self.local_ca = CA.LocalCA(self.lib_conf["ca"]["local_ca"]["dir"])
            if not self.local_ca.cert.exists():
                raise exceptions.SCAutolibMissingCA(
                    f"CA certificate not found in {str(self.local_ca.cert)}")
        else:
            self.ipa_ca = CA.IPAServerCA(self.lib_conf["ca"]["ipa"])
            if not self.ipa_ca.is_installed:
                raise exceptions.SCAutolibMissingCA(
                    "IPA server CA is not installed")
        logger.info("CA is initialized")



"""
Implementation of CLI commands for SCAutolib.
"""

import click
from pathlib import Path
from sys import exit

from SCAutolib import logger, exceptions, schema_user, ReturnCode
from SCAutolib.controller import Controller


@click.group()
@click.option("--conf", "-c",
              default="./conf.json",
              type=click.Path(exists=True, resolve_path=True),
              show_default=True,
              help="Path to JSON configuration file.")
@click.option('--force', "-f", is_flag=True, default=False, show_default=True,
              help="Force the command to overwrite configuration if it exists.")
@click.option("--verbose", "-v", default="DEBUG", show_default=True,
              type=click.Choice(
                  ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                  case_sensitive=False),
              help="Verbosity level.")
@click.pass_context
def cli(ctx, force, verbose, conf):
    logger.setLevel(verbose)
    ctx.ensure_object(dict)  # Create a dict to store the context
    ctx.obj["FORCE"] = force  # Store the force option in the context
    ctx.obj["CONTROLLER"] = Controller(conf)


@click.command()
@click.option("--ca-type", "-t",
              required=False,
              default='all',
              type=click.Choice(['all', 'local', 'ipa'], case_sensitive=False),
              show_default=True,
              help="Type of the CA to be configured. If not set, all CA's "
                   "from the config file would be configured")
@click.pass_context
def setup_ca(ctx, ca_type):
    """
    Configure the CA's in the config file. If more than one CA is
    specified, specified CA type would be configured.
    """
    cnt = ctx.obj["CONTROLLER"]
    if ca_type == 'all':
        cnt.setup_local_ca(force=ctx.obj["FORCE"])
        cnt.setup_ipa_client(force=ctx.obj["FORCE"])
    elif ca_type == 'local':
        cnt.setup_local_ca(force=ctx.obj["FORCE"])
    elif ca_type == 'ipa':
        cnt.setup_ipa_client(force=ctx.obj["FORCE"])
    exit(ReturnCode.SUCCESS.value)


@click.command()
@click.option("--gdm", "-g",
              required=False,
              default=False,
              is_flag=True,
              help="Install GDM package")
@click.option("--install-missing", "-i",
              required=False,
              default=False,
              is_flag=True,
              help="Install missing packages")
@click.pass_context
def prepare(ctx, gdm, install_missing):
    """Configure entire system for smart cards based on the config file."""
    ctx.obj["CONTROLLER"].prepare(ctx.obj["FORCE"], gdm, install_missing)
    exit(ReturnCode.SUCCESS.value)


@click.command()
@click.argument("name",
                required=True,
                default=None)
@click.option("--card-dir", "-d",
              required=False,
              default=None,
              help="Path to the directory where smart card should be created")
@click.option("--card-type", "-t",
              required=False,
              default="virtual",
              type=click.Choice(
                  ["virtual", "real", "removinator"], case_sensitive=False),
              show_default=True,
              help="Type of the smart card to be created")
@click.option("--passwd", "-p",
              required=False,
              default=None,
              show_default=True,
              help="Password for the user")
@click.option("--pin", "-P",
              required=False,
              default=None,
              show_default=True,
              help="PIN for the smart card")
@click.option("--user-type", "-T",
              required=False,
              default="local",
              type=click.Choice(["local", "ipa"], case_sensitive=False),
              show_default=True,
              help="Type of the user to be created")
@click.pass_context
def setup_user(ctx, name, card_dir, card_type, passwd, pin, user_type):
    """Configure user with smart cards (if set) based on the config file."""
    cnt = ctx.obj["CONTROLLER"]
    logger.info(f"Start setup user {name}")
    try:
        user_dict = cnt.get_user_dict(name)
    except exceptions.SCAutolibMissingUserConfig:
        logger.warning(f"User {name} not found in config file, "
                       f"trying to create a new one")
        if not all([card_dir, card_type, passwd, pin, user_type]):
            logger.error("Not all required arguments are set")
            logger.error("Required arguments: --card-dir, --pin, --password")
            exit(ReturnCode.ERROR.value)
        user_dict = schema_user.validate(
            {"name": name,
             "card_dir": Path(card_dir),
             "card_type": card_type,
             "passwd": passwd,
             "pin": pin,
             "local": user_type == "local"})
        logger.debug(f"User dict: {user_dict}")

    try:
        cnt.init_ca(user_dict["local"])
    except exceptions.SCAutolibMissingCA:
        logger.error("CA is not configured on the system")
        exit(ReturnCode.MISSING_CA.value)

    try:
        user = cnt.setup_user(user_dict, ctx.obj["FORCE"])
        cnt.enroll_card(user, ctx.obj["FORCE"])
    except exceptions.SCAutolibException:
        logger.error("Something went wrong")
        exit(ReturnCode.FAILURE.value)
    exit(ReturnCode.SUCCESS.value)


cli.add_command(setup_ca)
cli.add_command(prepare)
cli.add_command(setup_user)



"""
This module implements classes for communication with different types of cards
that we are using in the library. Those types are: virtual smart card, real
(physical) smart card in standard reader, cards in the removinator.
"""
import json
import re
import time
from pathlib import Path
from traceback import format_exc

from SCAutolib import run, logger, TEMPLATES_DIR, LIB_DUMP_CARDS
from SCAutolib.exceptions import SCAutolibException


class Card:
    """
    Interface for child classes. All child classes will rewrite common methods
    based on the type of the card.
    """
    uri: str = None
    dump_file: Path = None
    type: str = None
    _user = None
    _pattern: str = None

    def _set_uri(self):
        """
        Sets URI for given smart card. Uri is matched from ``p11tool`` command
        with regular expression. If URI is not matched, assertion exception
        would be raised

        :raise: AssertionError
        """
        cmd = ["p11tool", "--list-token-urls"]
        out = run(cmd).stdout
        urls = re.findall(self._pattern, out)
        assert len(urls) == 1, f"Card URI is not matched. Matched URIs: {urls}"
        self.uri = urls[0]

    def insert(self):
        """
        Insert the card.
        """
        ...

    def remove(self):
        """
        Remove the card.
        """
        ...

    def enroll(self):
        """
        Enroll the card (upload a certificate and a key on it)
        """
        ...

    @staticmethod
    def load(json_file, **kwars):
        with json_file.open("r") as f:
            cnt = json.load(f)

        card = None
        if cnt["type"] == "virtual":
            assert "user" in kwars.keys(),\
                "No user is provided to load the card."
            card = VirtualCard(user=kwars["user"],
                               softhsm2_conf=Path(cnt["softhsm"]))
            card.uri = cnt["uri"]
        else:
            raise SCAutolibException(
                f"Unknown card type: {cnt['type']}")
        return card


class VirtualCard(Card):
    """
    This class provides method for manipulating with virtual smart card. Virtual
    smart card by itself is represented by the systemd service in the system.
    The card relates to some user, so providing the user is essential for
    correct functioning of methods for the virtual smart card.

    Card root directory has to be created before calling any method
    """

    _service_name: str = None
    _service_location: Path = None
    _softhsm2_conf: Path = None
    _nssdb: Path = None
    _template: Path = Path(TEMPLATES_DIR, "virt_cacard.service")
    _pattern = r"(pkcs11:model=PKCS%2315%20emulated;" \
               r"manufacturer=Common%20Access%20Card;serial=.*)"
    _inserted: bool = False
    type = "virtual"

    def __init__(self, user, softhsm2_conf: Path = None):
        """
        Initialise virtual smart card. Constructor of the base class is also
        used.

        :param user: User of this card
        :type user: SCAutolib.models.user.User
        :param softhsm2_conf: path to SoftHSM2 configuration file
        :type softhsm2_conf: pathlib.Path
        """

        self.user = user
        assert self.user.card_dir.exists(), "Card root directory doesn't exists"

        self.dump_file = LIB_DUMP_CARDS.joinpath(f"card-{user.username}.json")

        self._private_key = self.user.key
        self._cert = self.user.cert

        self._service_name = f"virt-sc-{self.user.username}"
        self._service_location = Path(
            f"/etc/systemd/system/{self._service_name}.service")
        self.dump_file = LIB_DUMP_CARDS.joinpath(
            f"card-{self._user.username}.json")
        self._softhsm2_conf = softhsm2_conf if softhsm2_conf \
            else Path("/home", self.user.username, "softhsm2.conf")

    def __call__(self, insert: bool):
        """
        Call method for virtual smart card. It would be used in the context
        manager.

        :param insert: True if the card should be inserted, False otherwise
        :type insert: bool
        """
        if insert:
            self.insert()
        return self.__enter__()

    def __enter__(self):
        """
        Start of context manager for virtual smart card. The card would be
        inserted if ``insert`` parameter in constructor is specified.

        :return: self
        """
        assert self._service_location.exists(), \
            "Service for virtual sc doesn't exists."
        return self

    def __exit__(self, exp_type, exp_value, exp_traceback):
        """
        End of context manager for virtual smart card. If any exception was
        raised in the current context, it would be logged as an error.

        :param exp_type: Type of the exception
        :param exp_value: Value for the exception
        :param exp_traceback: Traceback of the exception
        """
        if exp_type is not None:
            logger.error("Exception in virtual smart card context")
            logger.error(format_exc())
        if self._inserted:
            self.remove()

    def to_dict(self):
        return {
            "softhsm": str(self._softhsm2_conf),
            "type": "virtual",
            "uri": self.uri,
            "username": self.user.username
        }

    @property
    def softhsm2_conf(self):
        return self._softhsm2_conf

    @softhsm2_conf.setter
    def softhsm2_conf(self, conf: Path):
        assert conf.exists(), "File doesn't exist"
        self._softhsm2_conf = conf

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, system_user):
        self._user = system_user
        self._nssdb = self.user.card_dir.joinpath("db")

    @property
    def service_location(self):
        return self._service_location

    def insert(self):
        """
        Insert virtual smart card by starting the corresponding service.
        """
        cmd = ["systemctl", "start", self._service_name]
        out = run(cmd, check=True)
        time.sleep(2)  # to prevent error with fast restarting of the service
        logger.info(f"Smart card {self._service_name} is inserted")
        self._inserted = True
        return out

    def remove(self):
        """
        Remove the virtual card by stopping the service
        """
        cmd = ["systemctl", "stop", self._service_name]
        out = run(cmd, check=True)
        time.sleep(2)  # to prevent error with fast restarting of the service
        logger.info(f"Smart card {self._service_name} is removed")
        self._inserted = False
        return out

    def enroll(self):
        """
        Upload certificate and private key to the virtual smart card (upload to
        NSS database) with pkcs11-tool.
        """
        cmd = ["pkcs11-tool", "--module", "libsofthsm2.so", "--slot-index",
               '0', "-w", self._user.key, "-y", "privkey", "--label",
               f"'{self._user.username}'", "-p", self._user.pin, "--set-id", "0",
               "-d", "0"]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf})
        logger.debug(
            f"User key {self._user.key} is added to virtual smart card")

        cmd = ['pkcs11-tool', '--module', 'libsofthsm2.so', '--slot-index', "0",
               '-w', self._user.cert, '-y', 'cert', '-p', self._user.pin,
               '--label', f"'{self._user.username}'", '--set-id', "0", '-d', "0"]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf})
        logger.debug(
            f"User certificate {self._user.cert} is added to virtual smart card")

        # To get URI of the card, the card has to be inserted
        # Virtual smart card can't be started without a cert and a key uploaded
        # to it, so URI can be set only after uploading the cert and a key
        with self:
            self.insert()
            self._set_uri()

    def create(self):
        """
        Creates SoftHSM2 token and systemd service for virtual smart card.
        Directory for NSS database is created in this method as separate DB is
        required for each virtual card.
        """

        assert self._softhsm2_conf.exists(), \
            "Can't proceed, SoftHSM2 conf doesn't exist"

        self.user.card_dir.joinpath("tokens").mkdir(exist_ok=True)

        p11lib = "/usr/lib64/pkcs11/libsofthsm2.so"
        # Initialize SoftHSM2 token. An error would be raised if token with same
        # label would be created.
        cmd = ["softhsm2-util", "--init-token", "--free", "--label",
               self.user.username, "--so-pin", "12345678",
               "--pin", self.user.pin]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf}, check=True)
        logger.debug(
            f"SoftHSM token is initialized with label '{self.user.username}'")

        # Initialize NSS db
        self._nssdb = self.user.card_dir.joinpath("db")
        self._nssdb.mkdir(exist_ok=True)
        run(f"modutil -create -dbdir sql:{self._nssdb} -force", check=True)
        logger.debug(f"NSS database is initialized in {self._nssdb}")

        out = run(f"modutil -list -dbdir sql:{self._nssdb}")
        if "library name: p11-kit-proxy.so" not in out.stdout:
            run(["modutil", "-force", "-add", 'SoftHSM PKCS#11', "-dbdir",
                 f"sql:{self._nssdb}", "-libfile", p11lib])
            logger.debug("SoftHSM support is added to NSS database")

        # Create systemd service
        with self._template.open() as tmp:
            with self._service_location.open("w") as f:
                f.write(tmp.read().format(username=self.user.username,
                                          softhsm2_conf=self._softhsm2_conf,
                                          card_dir=self.user.card_dir))

        logger.debug(f"Service is created in {self._service_location}")
        run("systemctl daemon-reload")

        return self



import os
import pwd
import pytest

from SCAutolib.models.user import BaseUser
from SCAutolib.utils import dump_to_json


@pytest.mark.skipif(os.getuid() != 0, reason="Requires root privileges!")
def test_add_and_remove_local_user(local_user):
    local_user.add_user()
    assert pwd.getpwnam(local_user.username)

    local_user.delete_user()
    with pytest.raises(Exception):
        pwd.getpwnam(local_user.username)


def test_add_and_remove_key_cert_pair(local_user, local_ca_fixture):
    local_user.key = local_ca_fixture._ca_key
    local_user.cert = local_ca_fixture._ca_cert
    assert local_user.key == local_ca_fixture._ca_key
    assert local_user.cert == local_ca_fixture._ca_cert

    del local_user.key
    del local_user.cert

    assert local_user.cert is None
    assert local_user.key is None


def test_add_and_remove_cnf(local_user, local_ca_fixture):
    local_user.cnf = local_ca_fixture._ca_cnf
    assert local_user.cnf == local_ca_fixture._ca_cnf

    del local_user.cnf
    assert local_user.cnf is None


def test_dump_and_load_user(local_user):
    dump_to_json(local_user)

    user, card_file = BaseUser.load(local_user.dump_file)

    assert user.username == local_user.username
    assert user.pin == local_user.pin
    assert user.password == local_user.password
    assert user.card_dir == local_user.card_dir



import pytest
from configparser import ConfigParser
from subprocess import check_output

from SCAutolib.controller import Controller


@pytest.fixture()
def controller(dummy_config):
    return Controller(dummy_config)


def test_parse_config(dummy_config):
    """Test that configuration is parsed and validated properly."""
    cnt = Controller(dummy_config)

    assert cnt.conf_path.is_absolute()
    assert isinstance(cnt.lib_conf, dict)


@pytest.mark.service_restart
def test_setup_system(controller):
    cnt: Controller = controller
    packages = ["opensc", "httpd", "sssd", "sssd-tools", "gnutls-utils",
                "pcsc-lite-ccid", "pcsc-lite", "virt_cacard", "vpcd",
                "softhsm"]

    check_output(["dnf", "remove", "softhsm", "-y"], encoding="utf8")

    cnt.setup_system(install_missing=True, gdm=False)

    for p in packages:
        out = check_output(["rpm", "-qa", p], encoding="utf-8")
        assert p in out
    current_sssd = ConfigParser()
    with open("/etc/sssd/sssd.conf", "r") as f:
        current_sssd.read_file(f)

    msg = "Current SSSD conf content is different from config parser object " \
          "in the controller"
    assert set(current_sssd.sections()).issubset(set(
        cnt.sssd_conf._default_parser.sections())), msg


@pytest.mark.ipa
def test_users_create_and_delete(controller, tmp_path, ipa_fixture):
    """Test for adding local and IPA users to the systems and initializing all
    required files."""
    cnt: Controller = controller
    cnt.ipa_ca = ipa_fixture

    try:
        for u in cnt.lib_conf["users"]:
            cnt.setup_user(u)
        for p in [t["card_dir"] for t in cnt.lib_conf["users"]]:
            assert p.joinpath("sofhtsm2.conf").exists()
    finally:
        for u in cnt.users:
            u.delete_user()



"""
This module provides a set of additional helping functions that are used
across the library. These functions are made based on library demands and are
not attended to cover some general use-cases or specific corner cases.
"""
import json
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from enum import Enum
from pathlib import Path

from SCAutolib import run, logger, TEMPLATES_DIR, LIB_DUMP_USERS, LIB_DUMP_CAS
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.models.CA import LocalCA, BaseCA, IPAServerCA
from SCAutolib.models.card import Card
from SCAutolib.models.file import OpensslCnf
from SCAutolib.models.user import BaseUser


class OSVersion(Enum):
    """
    Enumeration for Linux versions. Used for more convenient checks.
    """
    Fedora = 1
    RHEL_9 = 2
    RHEL_8 = 3
    CentOS_8 = 4
    CentOS_9 = 5


def _check_selinux():
    """
    Checks if specific SELinux module for virtual smart card is installed.
    This is implemented be checking the hardcoded name for the module
    (virtcacard) to be present in the list of SELinux modules. If this name is
    not present in the list, then virtcacard.cil file would be created in conf
    or subdirectory in the CA directory specified by the configuration file.
    """
    result = run("semodule -l", print_=False)
    if "virtcacard" not in result.stdout:
        logger.debug(
            "SELinux module for virtual smart cards is not present in the "
            "system. Installing...")

        run(["semodule", "-i", f"{TEMPLATES_DIR}/virtcacard.cil"])

        run(["systemctl", "restart", "pcscd"])
        logger.debug("pcscd service is restarted")

    logger.debug(
        "SELinux module for virtual smart cards is installed")


def _gen_private_key(key_path: Path):
    """
    Generate RSA private key to specified location.

    :param key_path: path to output certificate
    """
    # CAC specification do not specify key size specifies key size
    # up to 2048 bits, so keys greater than 2048 bits is not supported
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    with key_path.open("wb") as f:
        f.write(key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()))


def _get_os_version():
    """
    Find Linux version. Available version: RHEL 8, RHEL 9, Fedora.
    :return: Enum with OS version
    """
    with open('/etc/redhat-release', "r") as f:
        cnt = f.read()

    if "Red Hat Enterprise Linux release 9" in cnt:
        return OSVersion.RHEL_9
    elif "Red Hat Enterprise Linux release 8" in cnt:
        return OSVersion.RHEL_8
    elif "Fedora" in cnt:
        return OSVersion.Fedora
    elif "CentOS Stream release 8" in cnt:
        return OSVersion.CentOS_8
    elif "CentOS Stream release 9" in cnt:
        return OSVersion.CentOS_9
    else:
        raise SCAutolibException("OS is not detected.")


def _install_packages(packages):
    """
    Install given packages and log package version

    :param packages: list of packages to be installed
    """
    run(f"dnf install -y {' '.join(packages)}")
    for pkg in packages:
        pkg = run(["rpm", "-q", pkg]).stdout
        logger.debug(f"Package {pkg} is installed")


def _check_packages(packages):
    """
    Find missing packages

    :param packages: list of required packages
    :type packages: list
    :return: list of missing packages
    """
    missing = []
    for pkg in packages:
        # Return code 1 means the package is not installed
        out = run(["rpm", "-q", pkg], return_code=[0, 1])
        if out.returncode == 1:
            logger.warning(f"Package {pkg} is required for the testing, "
                           f"but is not present in the system")
            missing.append(pkg)
        else:
            logger.debug(f"Package {out.stdout.strip()} is present")
    return missing


def dump_to_json(obj):
    """
    Store serialised object to the JSON file.
    """
    with obj.dump_file.open("w") as f:
        json.dump(obj.to_dict(), f)
    logger.debug(f"Object {type(obj)} is stored to the {obj.dump_file} file")


def user_factory(username, **kwargs):
    """
    Load user with given username from JSON file. If user have the card file
    linked, then load it as well.

    :param username: username of the user
    :type username: str

    :return: user object
    :rtype: BaseUser
    """
    user_file = LIB_DUMP_USERS.joinpath(f"{username}.json")
    logger.debug(f"Loading user {username} from {user_file}")
    result = None
    user = None
    if user_file.exists():
        result = BaseUser.load(user_file, **kwargs)
    if type(result) == tuple:
        user, card_file = result
        logger.debug(f"Loading card from {card_file}")
        user.card = Card.load(card_file, user=user)
    else:
        user = result
    return user


def ipa_factory():
    """
    Create a new IPAServerCA object.

    .. note: Creating new IPA server with CA is not supported.

    :return: object of IPAServerCA
    :rtype: SCAutolib.models.CA.IPAServerCA
    """
    json_file = LIB_DUMP_CAS.joinpath("ipa-server.json")
    if not json_file.exists():
        msg = "Dump file for ipa server CA is not present."
        logger.error(msg)
        logger.error("The reason for this is most likely that the system was "
                     "not configured for IPA client via SCAutolib")
        raise SCAutolibException(msg)
    ca = BaseCA.load(json_file)
    if not isinstance(ca, IPAServerCA):
        msg = "Values in dump file are not valid for IPA server, so the " \
              "object can't be created"
        logger.error(msg)
        raise SCAutolibException(msg)
    return ca


def local_ca_factory(path: Path = None, force: bool = False):
    """
    Create a new LocalCA object or load existing one if no path provided.

    :param path: path to the CA directory
    :type path: Path
    :param force: force creation of new CA
    :type force: bool
    :return: object of LocalCA
    :rtype: SCAutolib.models.CA.LocalCA
    """
    if not path:
        return BaseCA.load(LIB_DUMP_CAS.joinpath("local-ca.json"))

    path.mkdir(exist_ok=True, parents=True)
    cnf = OpensslCnf(path.joinpath("ca.cnf"), "CA", str(path))
    ca = LocalCA(root_dir=path, cnf=cnf)
    if force:
        logger.warning(f"Removing previous local CA in a directory {path}")
        ca.cleanup()
    ca.dump_file = path.joinpath("ca-dump.json")
    cnf.create()
    cnf.save()
    ca.setup()
    run(["systemctl", "restart", "sssd"], sleep=5)
    return ca
